package be.iminds.ilabt.jfed.experimenter_gui.util;

/**
 * Based on: ReplacingInputStream
 * Source: https://github.com/Inbot/inbot-utils/blob/master/src/main/java/io/inbot/utils/ReplacingInputStream.java
 *
 * license:
 * The MIT License (MIT)
 *
 * Copyright (c) 2015 Inbot
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * ReplacingInputStream
 */

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;

/**
 * Simple FilterInputStream that can replace occurrances of bytes with something else.
 */
public class RemovingInputStream extends FilterInputStream {

    // while matching, this is where the bytes go.
    int[] buf=null;
    int matchedIndex=0;
    int unbufferIndex=0;
    int replacedIndex=0;

    private final byte[] patternStart;
    private final byte patternEnd;
    private final int maxRemoveLen;
    private State state=State.NOT_MATCHED;

    // simple state machine for keeping track of what we are doing
    private enum State {
        NOT_MATCHED,
        MATCHING,
        REMOVING,
        UNBUFFER
    }

    /**
     * Removes occurrences of pattern in the input. Note: input is assumed to be UTF-8 encoded. If not the case use byte[] based pattern and replacement.
     * both patternStart and patternEnd itself are also removed.
     * @param in input
     * @param patternStart pattern to start removing (inclusive).
     * @param patternEnd the char to stop removing (inclusive).
     */
    public RemovingInputStream(InputStream in, String patternStart, char patternEnd, int maxRemoveLen) {
        this(in, patternStart.getBytes(StandardCharsets.UTF_8),
                (patternEnd+"").getBytes(StandardCharsets.UTF_8)[0],
                maxRemoveLen);
        assert (patternEnd+"").getBytes(StandardCharsets.UTF_8).length == 1;
    }

    /**
     * Removes occurrences of pattern in the input.
     * both patternStart and patternEnd itself are also removed.
     * @param in input
     * @param patternStart pattern to start removing (inclusive).
     * @param patternEnd the char to stop removing (inclusive).
     */
    public RemovingInputStream(InputStream in, byte[] patternStart, byte patternEnd, int maxRemoveLen) {
        super(in);
        assert patternStart != null;
        assert patternStart.length > 0 : "pattern length ("+patternStart.length+") should be > 0";
        this.patternStart = patternStart;
        this.patternEnd = patternEnd;
        this.maxRemoveLen = maxRemoveLen;
        buf = new int[maxRemoveLen+1];
    }

    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        // copy of parent logic; we need to call our own read() instead of super.read(), which delegates instead of calling our read
        if (b == null) {
            throw new NullPointerException();
        } else if (off < 0 || len < 0 || len > b.length - off) {
            throw new IndexOutOfBoundsException();
        } else if (len == 0) {
            return 0;
        }

        int c = read();
        if (c == -1) {
            return -1;
        }
        b[off] = (byte)c;

        int i = 1;
        try {
            for (; i < len ; i++) {
                c = read();
                if (c == -1) {
                    break;
                }
                b[off + i] = (byte)c;
            }
        } catch (IOException ee) {
        }
        return i;

    }

    @Override
    public int read(byte[] b) throws IOException {
        // call our own read
        return read(b, 0, b.length);
    }

    @Override
    public int read() throws IOException {
        // use a simple state machine to figure out what we are doing
        int next;
        switch (state) {
            case NOT_MATCHED:
                // we are not currently matching, replacing, or unbuffering
                next=super.read();
                if(patternStart[0] == next) {
                    // clear whatever was there
                    buf=new int[maxRemoveLen+1]; // clear whatever was there
                    // make sure we start at 0
                    matchedIndex=0;

                    buf[matchedIndex++]=next;
                    if(patternStart.length == 1) {
                        // edgecase when the pattern length is 1 we go straight to replacing
                        state=State.REMOVING;
                        // reset replace counter
                        replacedIndex=0;
                    } else {
                        // pattern of length 1
                        state=State.MATCHING;
                    }
                    // recurse to continue matching
                    return read();
                } else {
                    return next;
                }
            case MATCHING:
                // the previous bytes matched part of the pattern
                next=super.read();
                while (next != -1 && state == State.MATCHING) {
                    if (patternStart[matchedIndex] == next) {
                        buf[matchedIndex++] = next;
                        if (matchedIndex == patternStart.length) {
                            // we've found a full match!
                            state = State.REMOVING;
                        } else {
                            next=super.read();
                        }
                    } else {
                        // mismatch -> unbuffer
                        buf[matchedIndex++] = next;
                        state = State.UNBUFFER;
                        unbufferIndex = 0;
                    }
                }

                //recursive call to self!
                return read();
            case REMOVING:
                next=super.read();
                while (next != -1 && state == State.REMOVING) {
                    buf[matchedIndex++] = next;
                    if (next == patternEnd) {
                        //full match, remove all
                        state = State.NOT_MATCHED;
                    } else {
                        if (matchedIndex >= maxRemoveLen) {
                            //match too long, remove nothing.
                            state = State.UNBUFFER;
                            unbufferIndex = 0;
                        } else {
                            next=super.read();
                        }
                    }
                }

                //recursive call to self!
                return read();

            case UNBUFFER:
                // we partially matched the pattern before encountering a non matching byte
                // we need to serve up the buffered bytes before we go back to NOT_MATCHED
                next=buf[unbufferIndex++];
                if(unbufferIndex==matchedIndex) {
                    state=State.NOT_MATCHED;
                    matchedIndex=0;
                }
                return next;
            default:
                throw new IllegalStateException("no such state " + state);
        }
    }

    @Override
    public String toString() {
        return state.name() + " " + matchedIndex + " " + replacedIndex + " " + unbufferIndex;
    }

}